/* Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdlib.h>
#include <errno.h>
#include <glib.h>

#include "iotcon-types.h"
#include "ic.h"
#include "ic-list.h"
#include "ic-value.h"
#include "ic-utils.h"
#include "ic-representation.h"
#include "ic-attributes.h"

iotcon_attributes_h icl_attributes_ref(iotcon_attributes_h attributes)
{
	RETV_IF(NULL == attributes, NULL);
	RETV_IF(attributes->ref_count <= 0, NULL);

	attributes->ref_count++;

	return attributes;
}

int icl_attributes_create(iotcon_attributes_h *ret_attributes)
{
	FN_CALL;
	iotcon_attributes_h attributes = NULL;

	RETV_IF(NULL == ret_attributes, IOTCON_ERROR_INVALID_PARAMETER);

	attributes = calloc(1, sizeof(struct icl_attributes_s));
	if (NULL == attributes) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	attributes->hash_table = g_hash_table_new_full(g_str_hash, g_str_equal, free,
			icl_value_destroy);
	attributes->ref_count = 1;

	*ret_attributes = attributes;

	return IOTCON_ERROR_NONE;
}

int icl_attributes_destroy(iotcon_attributes_h attributes)
{
	FN_CALL;
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);

	attributes->ref_count--;

	if (0 == attributes->ref_count) {
		g_hash_table_destroy(attributes->hash_table);
		free(attributes);
	}
	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_int(iotcon_attributes_h attributes, const char *key, int val)
{
	FN_CALL;
	iotcon_value_h value;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_int(val);
	if (NULL == value) {
		ERR("icl_value_create_int(%d) Fail", val);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_bool(iotcon_attributes_h attributes, const char *key, bool val)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_bool(val);
	if (NULL == value) {
		ERR("icl_value_create_bool(%d) Fail", val);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_double(iotcon_attributes_h attributes, const char *key, double val)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_double(val);
	if (NULL == value) {
		ERR("icl_value_create_double(%f) Fail", val);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_str(iotcon_attributes_h attributes, const char *key, char *val)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_str(val);
	if (NULL == value) {
		ERR("icl_value_create_str(%s) Fail", val);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_byte_str(iotcon_attributes_h attributes, const char *key, unsigned char *val, int len)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(len <= 0, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_byte_str(val, len);
	if (NULL == value) {
		ERR("icl_value_create_byte_str() Fail");
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_null(iotcon_attributes_h attributes, const char *key)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_null();
	if (NULL == value) {
		ERR("icl_value_create_null() Fail");
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_list(iotcon_attributes_h attributes, const char *key, iotcon_list_h list)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == list, IOTCON_ERROR_INVALID_PARAMETER);


	value = icl_value_create_list(list);
	if (NULL == value) {
		ERR("icl_value_create_list() Fail");
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

int icl_attributes_add_attributes(iotcon_attributes_h attributes, const char *key, iotcon_attributes_h val)
{
	FN_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = icl_value_create_attributes(val);
	if (NULL == value) {
		ERR("icl_value_create_attributes(%p) Fail", val);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

static int _icl_attributes_set_value(iotcon_attributes_h attributes, const char *key, iotcon_value_h value)
{
	FN_CALL;
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == value, IOTCON_ERROR_INVALID_PARAMETER);

	g_hash_table_replace(attributes->hash_table, ic_utils_strdup(key), value);

	return IOTCON_ERROR_NONE;
}

static void _icl_attributes_clone_foreach(char *key, iotcon_value_h src_val, iotcon_attributes_h dest_attributes)
{
	FN_CALL;
	iotcon_value_h copied_val;

	copied_val = icl_value_clone(src_val);
	if (NULL == copied_val) {
		ERR("icl_value_clone() Fail");
		return;
	}

	_icl_attributes_set_value(dest_attributes, key, copied_val);
}

int icl_attributes_clone(iotcon_attributes_h attributes, iotcon_attributes_h *attributes_clone)
{
	FN_CALL;
	int ret;
	iotcon_attributes_h temp = NULL;

	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == attributes_clone, IOTCON_ERROR_INVALID_PARAMETER);

	if (attributes->hash_table) {
		ret = icl_attributes_create(&temp);
		if (IOTCON_ERROR_NONE != ret)
			return ret;

		g_hash_table_foreach(attributes->hash_table, (GHFunc)_icl_attributes_clone_foreach,
				temp);
	}

	*attributes_clone = temp;

	return IOTCON_ERROR_NONE;
}
