/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <errno.h>
#include <glib.h>
#include <octypes.h>
#include <ocstack.h>
#include <ocpayload.h>
#include <pinoxmcommon.h>
#include <system_info.h>
#include <vconf.h>

#include "iotcon.h"
#include "ic.h"
#include "ic-utils.h"
#include "ic-types.h"
#include "ic-presence.h"
#include "ic-remote-resource.h"
#include "ic-representation.h"
#include "ic-query.h"
#include "ic-request.h"
#include "ic-resource.h"
#include "ic-resource-interfaces.h"
#include "ic-resource-types.h"
#include "ic-response.h"
#include "ic-observers.h"
#include "ic-attributes.h"
#include "ic-lite-resource.h"
#include "ic-ioty.h"
#include "ic-ioty-parse.h"
#include "ic-ioty-convert.h"
#include "ic-ioty-ocprocess.h"
#include "ic-cbor.h"


typedef struct {
	iotcon_generated_pin_cb cb;
	void *user_data;
} icl_generated_pin_cb_container_s;

static const char *IC_SYSTEM_INFO_PLATFORM_VERSION = "http://tizen.org/feature/platform.version";
static const char *IC_SYSTEM_INFO_MANUF_NAME = "http://tizen.org/system/manufacturer";
static const char *IC_SYSTEM_INFO_MODEL_NAME = "http://tizen.org/system/model_name";
static const char *IC_SYSTEM_INFO_BUILD_STRING = "http://tizen.org/system/build.string";
static const char *IC_SYSTEM_INFO_TIZEN_ID = "http://tizen.org/system/tizenid";

static bool icl_state;
static bool icl_is_set_generated_pin_cb;
static char icl_svr_db_file[PATH_MAX];
static OCPersistentStorage icl_ioty_ps;
static GList *icl_generated_pin_cb_list;

GHashTable *resource_hash_table;
void icl_ioty_deinit(pthread_t thread)
{
	FN_CALL;
	int ret;

	icl_ioty_ocprocess_stop();
	ic_utils_cond_signal(IC_UTILS_COND_POLLING);

	ret = pthread_join(thread, NULL);
	if (0 != ret)
		ERR("pthread_join() Fail(%d)", ret);

	ic_utils_cond_polling_destroy();

	ic_utils_mutex_lock(IC_UTILS_MUTEX_IOTY);
	ret = OCStop();
	if (OC_STACK_OK != ret)
		ERR("OCStop() Fail(%d)", ret);
	else
		icl_state = false;

	ic_utils_mutex_unlock(IC_UTILS_MUTEX_IOTY);
}

static FILE* _icl_ioty_ps_fopen(const char *path, const char *mode)
{
	return fopen(icl_svr_db_file, mode);
}

int icl_ioty_set_persistent_storage(const char *file_path, bool is_pt)
{
	int ret;

	RETV_IF(NULL == file_path, IOTCON_ERROR_INVALID_PARAMETER);
	SECURE_DBG("file_path: [%s], is_pt: [%d]", file_path, is_pt);

	if (-1 == access(file_path, F_OK)) {
		INFO("Create file");
		if (true == is_pt) {
			ret = icl_cbor_create_pt_svr_db(file_path);
			if (IOTCON_ERROR_NONE != ret)
				return ret;
		} else {
			ret = icl_cbor_create_svr_db(file_path);
			if (IOTCON_ERROR_NONE != ret)
				return ret;
		}
	} else if (-1 == access(file_path, R_OK | W_OK)) {
		ERR("access() Fail(%d)", errno);
		return IOTCON_ERROR_PERMISSION_DENIED;
	}
	snprintf(icl_svr_db_file, sizeof(icl_svr_db_file), "%s", file_path);

	icl_ioty_ps.open = _icl_ioty_ps_fopen;
	icl_ioty_ps.read = fread;
	icl_ioty_ps.write = fwrite;
	icl_ioty_ps.close = fclose;
	icl_ioty_ps.unlink = unlink;

	ic_utils_mutex_lock(IC_UTILS_MUTEX_IOTY);
	ret = OCRegisterPersistentStorageHandler(&icl_ioty_ps);
	if (OC_STACK_OK != ret) {
		ERR("OCRegisterPersistentStorageHandler() Fail(%d)", ret);
		ret = IOTCON_ERROR_IOTIVITY;
	} else {
		icl_state = true;
		ret = IOTCON_ERROR_NONE;
	}
	ic_utils_mutex_unlock(IC_UTILS_MUTEX_IOTY);

	return ret;
}

/* LCOV_EXCL_START */
static void _generated_pin_cb(char *pin, size_t length)
{
	GList *node;
	icl_generated_pin_cb_container_s *container;

	if (NULL == pin) {
		ERR("Invalid Pin Number");
		return;
	}
	INFO("PIN : %s", pin);

	for (node = icl_generated_pin_cb_list; node; node = node->next) {
		container = node->data;
		if (container->cb)
			container->cb(pin, container->user_data);
	}
}
/* LCOV_EXCL_STOP */

static icl_generated_pin_cb_container_s* _find_generated_pin_cb(
		iotcon_generated_pin_cb cb)
{
	GList *node;
	icl_generated_pin_cb_container_s *container;

	for (node = icl_generated_pin_cb_list; node; node = node->next) {
		container = node->data;
		if (cb == container->cb)
			return container;
	}

	return NULL;
}

int icl_ioty_add_generated_pin_cb(iotcon_generated_pin_cb cb, void *user_data)
{
	icl_generated_pin_cb_container_s *container = NULL;
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	if (_find_generated_pin_cb(cb)) {
		ERR("This callback is already registered.");
		return IOTCON_ERROR_ALREADY;
	}

	container = calloc(1, sizeof(icl_generated_pin_cb_container_s));
	if (NULL == container) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	container->cb = cb;
	container->user_data = user_data;

	icl_generated_pin_cb_list = g_list_append(icl_generated_pin_cb_list, container);

	if (false == icl_is_set_generated_pin_cb) {
		SetGeneratePinCB(&_generated_pin_cb);
		icl_is_set_generated_pin_cb = true;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remove_generated_pin_cb(iotcon_generated_pin_cb cb)
{
	icl_generated_pin_cb_container_s *container;
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	container = _find_generated_pin_cb(cb);
	if (NULL == container) {
		ERR("This callback is not registered");
		return IOTCON_ERROR_NO_DATA;
	}

	if (1 > g_list_length(icl_generated_pin_cb_list)) {
		ERR("It is required at least one function");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	icl_generated_pin_cb_list = g_list_remove(icl_generated_pin_cb_list, container);

	free(container);

	return IOTCON_ERROR_NONE;
}

int icl_ioty_mutex_lock()
{
	ic_utils_mutex_lock(IC_UTILS_MUTEX_IOTY);

	if (false == icl_state) {
		ERR("IoTCon is not initialized");
		ic_utils_mutex_unlock(IC_UTILS_MUTEX_IOTY);
		return IOTCON_ERROR_IOTIVITY;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_init(pthread_t *out_thread)
{
	FN_CALL;
	int ret;
	pthread_attr_t attr;

	RETV_IF(NULL == out_thread, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}

	INFO("OCInit2() called..");
	ret = OCInit2(OC_CLIENT_SERVER, OC_DEFAULT_FLAGS, OC_DEFAULT_FLAGS,
		OC_ADAPTER_IP|OC_ADAPTER_TCP);

	icl_ioty_mutex_unlock();
	if (OC_STACK_OK != ret) {
		ERR("OCInit2() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	icl_ioty_ocprocess_start();

	pthread_attr_init(&attr);

#ifdef THREAD_STACK_SIZE
	size_t stacksize = 0;
	ret = pthread_attr_setstacksize(&attr, stacksize);
	if (0 != ret)
		ERR("pthread_attr_setstacksize() Fail(%d)", ret);
#endif

	ret = ic_utils_cond_polling_init();
	if (IOTCON_ERROR_NONE != ret)
		ERR("ic_utils_cond_polling_init() Fail(%d)", ret);

	ret = pthread_create(out_thread, &attr, icl_ioty_ocprocess_thread, NULL);
	if (0 != ret) {
		ERR("pthread_create() Fail(%d)", ret);
		return IOTCON_ERROR_SYSTEM;
	}

	pthread_attr_destroy(&attr);

	return IOTCON_ERROR_NONE;
}

/* LCOV_EXCL_START */
static gboolean _icl_ioty_timeout(gpointer user_data)
{
	FN_CALL;
	int ret;
	icl_cb_s *cb_info = user_data;

	RETV_IF(NULL == cb_info, G_SOURCE_REMOVE);
	cb_info->timeout = 0;

	if (cb_info->cb) {
		DBG("Call the callback : op [%d], handle [%p], cb [%p]", cb_info->op, cb_info->handle, cb_info->cb);
		switch (cb_info->op) {
		case ICL_FIND_RESOURCE:
			((iotcon_found_resource_cb)cb_info->cb)(NULL, IOTCON_ERROR_TIMEOUT,
				cb_info->user_data);
			break;
		case ICL_GET_DEVICE_INFO:
			((iotcon_device_info_cb)cb_info->cb)(NULL, IOTCON_ERROR_TIMEOUT,
				cb_info->user_data);
			break;
		case ICL_GET_PLATFORM_INFO:
			((iotcon_platform_info_cb)cb_info->cb)(NULL, IOTCON_ERROR_TIMEOUT,
				cb_info->user_data);
			break;
		default:
			ERR("Invalid operation(%d)", cb_info->op);
			return G_SOURCE_REMOVE;
		}
	}
	DBG("Call the callback done");
	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCCancel(cb_info->handle, OC_LOW_QOS, NULL, 0);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCCancel() Fail(%d)", ret);
		return G_SOURCE_REMOVE;
	}

	FN_END;
	return G_SOURCE_REMOVE;
}
/* LCOV_EXCL_STOP */

static gboolean _icl_ioty_free_cb_data_idle_cb(gpointer p)
{
	FN_CALL;

	icl_cb_s *cb_info = p;
	free(cb_info);
	return G_SOURCE_REMOVE;
}

static void _icl_ioty_free_cb_data(void *data)
{
	icl_cb_s *cb_info = data;
	RET_IF(NULL == cb_info);

	if (cb_info->timeout) {
		g_source_remove(cb_info->timeout);
		cb_info->timeout = 0;
	}
	g_idle_add(_icl_ioty_free_cb_data_idle_cb, cb_info);
}

static char* _icl_ioty_resource_generate_uri(const char *uri_path, iotcon_query_h query)
{
	int len;
	bool loop_first = true;
	char uri_buf[PATH_MAX] = {0};
	GHashTableIter iter;
	gpointer key, value;

	RETV_IF(NULL == uri_path, NULL);

	if (NULL == query || NULL == query->hash)
		return strdup(uri_path);

	len = snprintf(uri_buf, sizeof(uri_buf), "%s", uri_path);

	/* remove suffix '/' */
	if ('/' == uri_buf[strlen(uri_buf) - 1]) {
		uri_buf[strlen(uri_buf) - 1] = '\0';
		len--;
	}

	if (query->hash) {
		g_hash_table_iter_init(&iter, query->hash);
		while (g_hash_table_iter_next(&iter, &key, &value)) {
			int query_len;
			DBG("query exist. key(%s), value(%s)", (char *)key, (char *)value);

			if (true == loop_first) {
				query_len = snprintf(uri_buf + len, sizeof(uri_buf) - len, "?%s=%s",
						(char *)key, (char *)value);
				loop_first = false;
			} else {
				query_len = snprintf(uri_buf + len, sizeof(uri_buf) - len, ";%s=%s",
						(char *)key, (char *)value);
			}
			len += query_len;
		}
	}

	return strdup(uri_buf);
}

int icl_ioty_find_resource(const char *host_address,
		int connectivity_type,
		iotcon_query_h query,
		iotcon_found_resource_cb cb,
		void *user_data)
{
	int conn_options;
	int ret, timeout;
	char *coap_str = NULL;
	char *full_uri;
	char uri[PATH_MAX] = {0};
	icl_cb_s *cb_data;
	OCDoHandle handle = NULL;
	OCCallbackData cbdata = {0};
	OCConnectivityType oic_conn_type;

	SECURE_DBG("host_address:[%s], connectivity_type:[0x%x]", host_address, connectivity_type);
	if (host_address	&& IC_EQUAL != strncmp(IC_COAP_PREFIX, host_address, strlen(IC_COAP_PREFIX)))
		coap_str = IC_COAPS;

	if (NULL == host_address) {
		snprintf(uri, sizeof(uri), "%s", OC_RSRVD_WELL_KNOWN_URI);
	} else if (NULL == coap_str) {
		snprintf(uri, sizeof(uri), "%s%s", host_address, OC_RSRVD_WELL_KNOWN_URI);
	} else {
		snprintf(uri, sizeof(uri), "%s%s%s", coap_str, host_address,
				OC_RSRVD_WELL_KNOWN_URI);
	}

	full_uri = _icl_ioty_resource_generate_uri(uri, query);
	if (NULL == full_uri) {
		ERR("_icl_ioty_resource_generate_uri() Fail");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	cb_data = calloc(1, sizeof(icl_cb_s));
	if (NULL == cb_data) {
		ERR("calloc() Fail(%d)", errno);
		free(full_uri);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	cb_data->op = ICL_FIND_RESOURCE;
	cb_data->cb = cb;
	cb_data->user_data = user_data;
	cb_data->connectivity_type = connectivity_type % IC_CONNECTIVITY_REMOVE_EXTRA_OPTIONS;
	cb_data->connectivity_extra = connectivity_type - cb_data->connectivity_type;

	cbdata.context = cb_data;
	cbdata.cb = icl_ioty_ocprocess_find_cb;
	cbdata.cd = _icl_ioty_free_cb_data;

	conn_options = ic_utils_host_address_get_connectivity(host_address,
			cb_data->connectivity_type);
	if (NULL == host_address && IOTCON_CONNECTIVITY_IP == cb_data->connectivity_type) {
		if (IOTCON_CONNECTIVITY_IPV4_ONLY & cb_data->connectivity_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV4;
		else if (IOTCON_CONNECTIVITY_IPV6_ONLY & cb_data->connectivity_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV6;
	}
	oic_conn_type = ic_ioty_convert_connectivity_type(cb_data->connectivity_type,
			conn_options);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		free(full_uri);
		return ret;
	}
	SECURE_DBG("full_uri:[%s], oic_conn_type:[0x%x]", full_uri, oic_conn_type);
	ret = OCDoResource(&handle, OC_REST_DISCOVER, full_uri, NULL, NULL, oic_conn_type,
			OC_LOW_QOS, &cbdata, NULL, 0);
	free(full_uri);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(DISCOVER) Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		return ic_ioty_parse_oic_error(ret);
	}

	cb_data->handle = handle;
	DBG("handle [%p], cb [%p]", cb_data->handle, cb_data->cb);

	icl_get_timeout(&timeout);
	cb_data->timeout = g_timeout_add_seconds(timeout, _icl_ioty_timeout, cb_data);

	return IOTCON_ERROR_NONE;
}

int icl_ioty_find_device_info(const char *host_address,
		int connectivity_type,
		iotcon_query_h query,
		iotcon_device_info_cb cb,
		void *user_data)
{
	int ret, timeout;
	int conn_extra;
	int conn_options;
	char *coap_str = NULL;
	char *full_uri;
	char uri[PATH_MAX] = {0};
	icl_cb_s *cb_data = NULL;
	OCDoHandle handle = NULL;
	OCCallbackData cbdata = {0};
	OCConnectivityType oic_conn_type;

	SECURE_DBG("host_address:[%s], connectivity_type:[0x%x]", host_address, connectivity_type);
	if (host_address	&& IC_EQUAL != strncmp(IC_COAP_PREFIX, host_address, strlen(IC_COAP_PREFIX)))
		coap_str = IC_COAP;

	if (NULL == host_address)
		snprintf(uri, sizeof(uri), "%s", OC_RSRVD_DEVICE_URI);
	else if (NULL == coap_str)
		snprintf(uri, sizeof(uri), "%s%s", host_address, OC_RSRVD_DEVICE_URI);
	else
		snprintf(uri, sizeof(uri), "%s%s%s", coap_str, host_address, OC_RSRVD_DEVICE_URI);

	full_uri = _icl_ioty_resource_generate_uri(uri, query);
	if (NULL == full_uri) {
		ERR("_icl_ioty_resource_generate_uri() Fail");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	cb_data = calloc(1, sizeof(icl_cb_s));
	if (NULL == cb_data) {
		ERR("calloc() Fail(%d)", errno);
		free(full_uri);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	cb_data->cb = cb;
	cb_data->user_data = user_data;
	cb_data->op = ICL_GET_DEVICE_INFO;

	cbdata.context = cb_data;
	cbdata.cb = icl_ioty_ocprocess_device_info_cb;
	cbdata.cd = _icl_ioty_free_cb_data;

	conn_extra = connectivity_type;
	connectivity_type %= IC_CONNECTIVITY_REMOVE_EXTRA_OPTIONS;
	conn_extra -= connectivity_type;

	conn_options = ic_utils_host_address_get_connectivity(host_address, connectivity_type);
	if (NULL == host_address && IOTCON_CONNECTIVITY_IP == connectivity_type) {
		if (IOTCON_CONNECTIVITY_IPV4_ONLY & conn_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV4;
		else if (IOTCON_CONNECTIVITY_IPV6_ONLY & conn_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV6;
	}
	oic_conn_type = ic_ioty_convert_connectivity_type(connectivity_type, conn_options);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		free(full_uri);
		return ret;
	}
	SECURE_DBG("full_uri:[%s], oic_conn_type:[0x%x]", full_uri, oic_conn_type);
	ret = OCDoResource(&handle, OC_REST_DISCOVER, full_uri, NULL, NULL, oic_conn_type,
			OC_LOW_QOS, &cbdata, NULL, 0);
	free(full_uri);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(DISCOVER) Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		return ic_ioty_parse_oic_error(ret);
	}

	cb_data->handle = handle;
	DBG("handle [%p], cb [%p]", cb_data->handle, cb_data->cb);

	icl_get_timeout(&timeout);
	cb_data->timeout = g_timeout_add_seconds(timeout, _icl_ioty_timeout, cb_data);

	return IOTCON_ERROR_NONE;
}

int icl_ioty_find_platform_info(const char *host_address,
		int connectivity_type,
		iotcon_query_h query,
		iotcon_platform_info_cb cb,
		void *user_data)
{
	int ret, timeout;
	int conn_extra;
	int conn_options;
	char *coap_str = NULL;
	char *full_uri;
	char uri[PATH_MAX] = {0};
	icl_cb_s *cb_data = NULL;
	OCDoHandle handle = NULL;
	OCCallbackData cbdata = {0};
	OCConnectivityType oic_conn_type;

	SECURE_DBG("host_address:[%s], connectivity_type:[0x%x]", host_address, connectivity_type);
	if (host_address	&& IC_EQUAL != strncmp(IC_COAP_PREFIX, host_address, strlen(IC_COAP_PREFIX)))
		coap_str = IC_COAP;

	if (NULL == host_address)
		snprintf(uri, sizeof(uri), "%s", OC_RSRVD_PLATFORM_URI);
	else if (NULL == coap_str)
		snprintf(uri, sizeof(uri), "%s%s", host_address, OC_RSRVD_PLATFORM_URI);
	else
		snprintf(uri, sizeof(uri), "%s%s%s", coap_str, host_address, OC_RSRVD_PLATFORM_URI);

	full_uri = _icl_ioty_resource_generate_uri(uri, query);
	if (NULL == full_uri) {
		ERR("_icl_ioty_resource_generate_uri() Fail");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	cb_data = calloc(1, sizeof(icl_cb_s));
	if (NULL == cb_data) {
		ERR("calloc() Fail(%d)", errno);
		free(full_uri);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	cb_data->cb = cb;
	cb_data->user_data = user_data;
	cb_data->op = ICL_GET_PLATFORM_INFO;

	cbdata.context = cb_data;
	cbdata.cb = icl_ioty_ocprocess_platform_info_cb;
	cbdata.cd = _icl_ioty_free_cb_data;

	conn_extra = connectivity_type;
	connectivity_type %= IC_CONNECTIVITY_REMOVE_EXTRA_OPTIONS;
	conn_extra -= connectivity_type;

	conn_options = ic_utils_host_address_get_connectivity(host_address, connectivity_type);
	if (NULL == host_address && IOTCON_CONNECTIVITY_IP == connectivity_type) {
		if (IOTCON_CONNECTIVITY_IPV4_ONLY & conn_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV4;
		else if (IOTCON_CONNECTIVITY_IPV6_ONLY & conn_extra)
			conn_options |= IC_UTILS_CONNECTIVITY_IPV6;
	}
	oic_conn_type = ic_ioty_convert_connectivity_type(connectivity_type, conn_options);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		free(full_uri);
		return ret;
	}
	SECURE_DBG("full_uri:[%s], oic_conn_type:[0x%x]", full_uri, oic_conn_type);
	ret = OCDoResource(&handle, OC_REST_DISCOVER, full_uri, NULL, NULL, oic_conn_type,
			OC_LOW_QOS, &cbdata, NULL, 0);
	free(full_uri);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(DISCOVER) Fail(%d)", ret);
		_icl_ioty_free_cb_data(cb_data);
		return ic_ioty_parse_oic_error(ret);
	}

	cb_data->handle = handle;
	DBG("handle [%p], cb [%p]", cb_data->handle, cb_data->cb);

	icl_get_timeout(&timeout);
	cb_data->timeout = g_timeout_add_seconds(timeout, _icl_ioty_timeout, cb_data);

	return IOTCON_ERROR_NONE;
}

int icl_ioty_set_device_info(const char *device_name)
{
	int ret;
	OCDeviceInfo device_info = {0};

	RETV_IF(NULL == device_name, IOTCON_ERROR_INVALID_PARAMETER);

	device_info.deviceName = strdup(device_name);
	if (NULL == device_info.deviceName) {
		ERR("strdup() Fail");
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		free(device_info.deviceName);
		return ret;
	}
	ret = OCSetDeviceInfo(device_info);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCSetDeviceInfo() Fail(%d)", ret);
		free(device_info.deviceName);
		return ic_ioty_parse_oic_error(ret);
	}
	free(device_info.deviceName);

	return IOTCON_ERROR_NONE;
}

static int _icl_ioty_get_platform_info(OCPlatformInfo *platform_info)
{
	int ret;
	char *tizen_id = NULL;
	char *device_name = NULL;
	char platform_id[1024];

	RETV_IF(NULL == platform_info, IOTCON_ERROR_INVALID_PARAMETER);

	device_name = vconf_get_str(VCONFKEY_SETAPPL_DEVICE_NAME_STR);
	WARN_IF(device_name == NULL, "vconf_get_str() Fail");

	ret = system_info_get_platform_string(IC_SYSTEM_INFO_TIZEN_ID, &tizen_id);
	WARN_IF(SYSTEM_INFO_ERROR_NONE != ret, "system_info_get_platform_string() Fail(%d)", ret);

	snprintf(platform_id, sizeof(platform_id), "%s(%s)", IC_SAFE_STR(device_name), IC_SAFE_STR(tizen_id));

	free(device_name);
	free(tizen_id);
	SECURE_DBG("platform_id: %s", platform_id);

	/* Mandatory (oic.wk.p) */
	platform_info->platformID = strdup(platform_id);

	/* Mandatory (oic.wk.p) */
	ret = system_info_get_platform_string(IC_SYSTEM_INFO_MANUF_NAME,
			&platform_info->manufacturerName);
	if (SYSTEM_INFO_ERROR_NONE != ret) {
		ERR("system_info_get_platform_string(manufacturer) Fail(%d)", ret);
		free(platform_info->platformID);
		return IOTCON_ERROR_SYSTEM;
	}

	ret = system_info_get_platform_string(IC_SYSTEM_INFO_MODEL_NAME,
			&platform_info->modelNumber);
	if (SYSTEM_INFO_ERROR_NONE != ret)
		WARN("system_info_get_platform_string(model_name) Fail(%d)", ret);

	ret = system_info_get_platform_string(IC_SYSTEM_INFO_PLATFORM_VERSION,
			&platform_info->platformVersion);
	if (SYSTEM_INFO_ERROR_NONE != ret)
		WARN("system_info_get_platform_string(platform_version) Fail(%d)", ret);

	ret = system_info_get_platform_string(IC_SYSTEM_INFO_BUILD_STRING,
			&platform_info->firmwareVersion);
	if (SYSTEM_INFO_ERROR_NONE != ret)
		WARN("system_info_get_platform_string(build_string) Fail(%d)", ret);

	/* platform_info.manufacturerUrl */
	/* platform_info.dateOfManufacture */
	/* platform_info.operatingSystemVersion */
	/* platform_info.hardwareVersion */
	/* platform_info.supportUrl */
	/* platform_info.systemTime */

	return IOTCON_ERROR_NONE;
}

static void _icl_ioty_free_platform_info(OCPlatformInfo *platform_info)
{
	RET_IF(NULL == platform_info);

	free(platform_info->platformID);
	free(platform_info->manufacturerName);
	free(platform_info->manufacturerUrl);
	free(platform_info->modelNumber);
	free(platform_info->dateOfManufacture);
	free(platform_info->platformVersion);
	free(platform_info->operatingSystemVersion);
	free(platform_info->hardwareVersion);
	free(platform_info->firmwareVersion);
	free(platform_info->supportUrl);
	free(platform_info->systemTime);
}

int icl_ioty_set_platform_info()
{
	int ret;
	OCPlatformInfo platform_info = {0};

	ret = _icl_ioty_get_platform_info(&platform_info);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_get_platform_info() Fail(%d)", ret);
		return ret;
	}

	ic_utils_mutex_lock(IC_UTILS_MUTEX_IOTY);
	ret = OCSetPlatformInfo(platform_info);
	_icl_ioty_free_platform_info(&platform_info);
	ic_utils_mutex_unlock(IC_UTILS_MUTEX_IOTY);
	if (OC_STACK_OK != ret) {
		ERR("OCSetPlatformInfo() Fail(%d)", ret);
		ret = ic_ioty_parse_oic_error(ret);
	} else {
		ret = IOTCON_ERROR_NONE;
	}

	return ret;
}

int icl_ioty_add_presence_cb(const char *host_address,
		iotcon_connectivity_type_e connectivity_type,
		const char *resource_type,
		iotcon_presence_cb cb,
		void *user_data,
		iotcon_presence_h *presence_handle)
{
	int ret;
	OCDoHandle handle = NULL;
	char uri[PATH_MAX] = {0};
	OCCallbackData cbdata = {0};
	OCConnectivityType oic_conn_type = CT_ADAPTER_IP;
	iotcon_presence_h presence = NULL;
	OCDevAddr dev_addr = {0};

	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == presence_handle, IOTCON_ERROR_INVALID_PARAMETER);

	snprintf(uri, sizeof(uri), "%s", OC_RSRVD_PRESENCE_URI);

	presence = calloc(1, sizeof(struct icl_presence));
	if (NULL == presence) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	presence->cb = cb;
	presence->user_data = user_data;
	if (host_address)
		presence->host_address = strdup(host_address);
	if (resource_type)
		presence->resource_type = strdup(resource_type);
	if (!(IOTCON_CONNECTIVITY_IP & connectivity_type)) {
		DBG("default IP type is addeded");
		connectivity_type |= IOTCON_CONNECTIVITY_IP;
	}
	presence->connectivity_type = connectivity_type;

	if (IOTCON_CONNECTIVITY_PREFER_TCP & connectivity_type)
		oic_conn_type |= CT_ADAPTER_TCP;
	if (IOTCON_CONNECTIVITY_IPV4_ONLY & connectivity_type)
		oic_conn_type |= CT_IP_USE_V4;
	if (IOTCON_CONNECTIVITY_IPV6_ONLY & connectivity_type)
		oic_conn_type |= CT_IP_USE_V6;

	cbdata.context = presence;
	cbdata.cd = NULL;
	cbdata.cb = icl_ioty_ocprocess_presence_cb;

	if (NULL != host_address) {
		int temp_conn_type;
		int conn_options;
		temp_conn_type = connectivity_type % IC_CONNECTIVITY_REMOVE_EXTRA_OPTIONS;
		conn_options = ic_utils_host_address_get_connectivity(host_address, temp_conn_type);
		ret = ic_ioty_convert_host_address(host_address,
			temp_conn_type, conn_options, &dev_addr);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("ic_ioty_convert_host_address() Fail(%d)", ret);
			icl_destroy_presence(presence);
			return ret;
		}
	}

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		icl_destroy_presence(presence);
		return ret;
	}
	SECURE_DBG("uri:[%s], oic_conn_type:[0x%x]", uri, oic_conn_type);
	if (NULL != host_address) {
		DBG("dev_addr.addr:[%s], dev_addr.port:[%d]", dev_addr.addr, dev_addr.port);
		ret = OCDoResource(&handle, OC_REST_PRESENCE, uri, &dev_addr, NULL, oic_conn_type,
			OC_LOW_QOS, &cbdata, NULL, 0);
	} else {
		ret = OCDoResource(&handle, OC_REST_PRESENCE, uri, NULL, NULL, oic_conn_type,
			OC_LOW_QOS, &cbdata, NULL, 0);
	}
	DBG("handle:[%p]", handle);

	icl_ioty_mutex_unlock();
	presence->handle = handle;

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(PRESENCE) Fail(%d)", ret);
		icl_destroy_presence(presence);
		return IOTCON_ERROR_IOTIVITY;
	}

	*presence_handle = presence;

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remove_presence_cb(iotcon_presence_h presence)
{
	int ret;

	RETV_IF(NULL == presence, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}

	DBG("presence_handle->handle:[%p]", presence->handle);
	ret = OCCancel(presence->handle, OC_LOW_QOS, NULL, 0);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCCancel() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	icl_destroy_presence(presence);

	return IOTCON_ERROR_NONE;
}

static void _icl_ioty_free_observe_container(void *data)
{
	icl_observe_container_s *cb_container = data;

	RET_IF(NULL == cb_container);

	icl_remote_resource_unref(cb_container->resource);
	free(cb_container);
}


static int _icl_ioty_remote_resource_observe(iotcon_remote_resource_h resource,
		iotcon_observe_policy_e observe_policy,
		iotcon_query_h query,
		iotcon_remote_resource_observe_cb cb,
		void *user_data,
		OCDoHandle *obs_handle)
{
	int ret, options_size = 0;
	char *uri = NULL;
	OCMethod method;
	OCDevAddr dev_addr = {0};
	OCCallbackData cbdata = {0};
	OCConnectivityType oic_conn_type;
	OCHeaderOption *oic_options_ptr = NULL;
	OCHeaderOption oic_options[MAX_HEADER_OPTIONS];
	icl_observe_container_s *cb_container;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	/* uri */
	uri = _icl_ioty_resource_generate_uri(resource->uri_path, query);
	if (NULL == uri) {
		ERR("_icl_ioty_resource_generate_uri() Fail");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	/* method */
	switch (observe_policy) {
	case IOTCON_OBSERVE_IGNORE_OUT_OF_ORDER:
		method = OC_REST_OBSERVE;
		break;
	case IOTCON_OBSERVE_ACCEPT_OUT_OF_ORDER:
	default:
		method = OC_REST_OBSERVE_ALL;
	}

	/* connectivity type */
	oic_conn_type = ic_ioty_convert_connectivity_type(resource->connectivity_type,
			resource->connectivity_options);

	/* host address, port */
	ret = ic_ioty_convert_host_address(resource->host_address,
			resource->connectivity_type, resource->connectivity_options, &dev_addr);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("ic_ioty_convert_host_address() Fail(%d)", ret);
		free(uri);
		return ret;
	}

	if (IOTCON_RESOURCE_SECURE & resource->policies)
		dev_addr.flags |= OC_FLAG_SECURE;

	/* options */
	if (resource->header_options && resource->header_options->hash) {
		ret = ic_ioty_convert_header_options(resource->header_options, oic_options,
				sizeof(oic_options)/sizeof(oic_options[0]));
		if (IOTCON_ERROR_NONE != ret) {
			ERR("ic_ioty_convert_header_options() Fail(%d)", ret);
			free(uri);
			return ret;
		}
		oic_options_ptr = oic_options;
		options_size = g_hash_table_size(resource->header_options->hash);
	}

	cb_container = calloc(1, sizeof(icl_observe_container_s));
	if (NULL == cb_container) {
		ERR("calloc() Fail(%d)", errno);
		free(uri);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	cb_container->cb = cb;
	cb_container->user_data = user_data;
	cb_container->resource = resource;
	icl_remote_resource_ref(resource);

	cbdata.context = cb_container;
	cbdata.cb = icl_ioty_ocprocess_observe_cb;
	cbdata.cd = _icl_ioty_free_observe_container;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		_icl_ioty_free_observe_container(cb_container);
		free(uri);
		return ret;
	}

	SECURE_DBG("uri:[%s], oic_conn_type:[0x%x]", uri, oic_conn_type);
	ret = OCDoResource(obs_handle, method, uri, &dev_addr, NULL, oic_conn_type,
			OC_HIGH_QOS, &cbdata, oic_options_ptr, options_size);
	free(uri);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(OBSERVE:%d) Fail(%d)", method, ret);
		_icl_ioty_free_observe_container(cb_container);
		return IOTCON_ERROR_IOTIVITY;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_observe_register(
		iotcon_remote_resource_h resource,
		iotcon_observe_policy_e observe_policy,
		iotcon_query_h query,
		iotcon_remote_resource_observe_cb cb,
		void *user_data)
{
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(resource->obs_handle, IOTCON_ERROR_ALREADY);

	ret = _icl_ioty_remote_resource_observe(resource, observe_policy, query, cb, user_data,
			&(resource->obs_handle));
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_observe() Fail(%d)", ret);
		return ret;
	}
	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_observe_cancel(iotcon_remote_resource_h resource,
		OCDoHandle obs_handle)
{
	int ret;
	int options_size = 0;
	OCHeaderOption *oic_options_ptr = NULL;
	OCHeaderOption oic_options[MAX_HEADER_OPTIONS];

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	/* options */
	if (resource->header_options && resource->header_options->hash) {
		ret = ic_ioty_convert_header_options(resource->header_options, oic_options,
				sizeof(oic_options)/sizeof(oic_options[0]));
		if (IOTCON_ERROR_NONE != ret) {
			ERR("ic_ioty_convert_header_options() Fail(%d)", ret);
			return ret;
		}
		oic_options_ptr = oic_options;
		options_size = g_hash_table_size(resource->header_options->hash);
	}

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCCancel(obs_handle, OC_HIGH_QOS, oic_options_ptr, options_size);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCCancel() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_observe_deregister(
		iotcon_remote_resource_h resource)
{
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource->obs_handle, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_remote_resource_observe_cancel(resource, resource->obs_handle);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_observe_cancel() Fail(%d)", ret);
		return ret;
	}

	resource->obs_handle = NULL;
	return IOTCON_ERROR_NONE;
}

static void _icl_ioty_free_response_container(void *data)
{
	icl_response_container_s *cb_container = data;

	RET_IF(NULL == cb_container);

	if (cb_container->timeout) {
		g_source_remove(cb_container->timeout);
		cb_container->timeout = 0;
	}

	icl_remote_resource_unref(cb_container->resource);
	g_main_context_unref(cb_container->thread_context);

	free(cb_container);
}

/* LCOV_EXCL_START */
static gboolean _icl_ioty_response_timeout(gpointer p)
{
	icl_response_container_s *cb_container = p;

	RETV_IF(NULL == cb_container, G_SOURCE_REMOVE);

	cb_container->timeout = 0;

	if (cb_container->cb) {
		cb_container->cb(cb_container->resource, IOTCON_ERROR_TIMEOUT,
				cb_container->req_type, NULL, cb_container->user_data);
	}

	return G_SOURCE_REMOVE;
}
/* LCOV_EXCL_STOP */

static int _icl_ioty_remote_resource_crud(
		iotcon_request_type_e req_type,
		iotcon_remote_resource_h resource,
		iotcon_representation_h repr,
		iotcon_query_h query,
		iotcon_remote_resource_response_cb cb,
		void *user_data)
{
	int ret, timeout, options_size = 0;
	char *uri;
	icl_response_container_s *cb_container;
	OCCallbackData cbdata = {0};
	OCHeaderOption *oic_options_ptr = NULL;
	OCHeaderOption oic_options[MAX_HEADER_OPTIONS];
	OCConnectivityType oic_conn_type;
	OCDevAddr dev_addr = {0};
	OCMethod method;
	OCRepPayload *payload = NULL;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	DBG("req_type [%d]", req_type);
	/* method (request type) */
	method = ic_ioty_convert_request_type(req_type);

	uri = _icl_ioty_resource_generate_uri(resource->uri_path, query);
	if (NULL == uri) {
		ERR("_icl_ioty_resource_generate_uri() Fail");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	/* host address, port */
	ret = ic_ioty_convert_host_address(resource->host_address,
			resource->connectivity_type, resource->connectivity_options, &dev_addr);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("ic_ioty_convert_host_address() Fail(%d)", ret);
		free(uri);
		return ret;
	}

	if (IOTCON_RESOURCE_SECURE & resource->policies)
		dev_addr.flags |= OC_FLAG_SECURE;

	/* representation */
	if (repr) {
		ret = ic_ioty_convert_representation(repr, &payload);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("ic_ioty_convert_representation() Fail(%d)", ret);
			free(uri);
			return ret;
		}
	}

	/* connectivity type */
	oic_conn_type = ic_ioty_convert_connectivity_type(resource->connectivity_type,
			resource->connectivity_options);

	/* options */
	if (resource->header_options && resource->header_options->hash) {
		ret = ic_ioty_convert_header_options(resource->header_options, oic_options,
				sizeof(oic_options)/sizeof(oic_options[0]));
		if (IOTCON_ERROR_NONE != ret) {
			ERR("ic_ioty_convert_header_options() Fail(%d)", ret);
			OCRepPayloadDestroy(payload);
			free(uri);
			return ret;
		}
		oic_options_ptr = oic_options;
		options_size = g_hash_table_size(resource->header_options->hash);
	}

	cb_container = calloc(1, sizeof(icl_response_container_s));
	if (NULL == cb_container) {
		ERR("calloc() Fail(%d)", errno);
		OCRepPayloadDestroy(payload);
		free(uri);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	cb_container->resource = resource;
	icl_remote_resource_ref(cb_container->resource);
	cb_container->req_type = req_type;
	cb_container->cb = cb;
	cb_container->user_data = user_data;
	cb_container->thread_context = g_main_context_ref_thread_default();

	cbdata.context = cb_container;
	cbdata.cb = icl_ioty_ocprocess_crud_cb;
	cbdata.cd = _icl_ioty_free_response_container;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		_icl_ioty_free_response_container(cb_container);
		free(uri);
		return ret;
	}
	SECURE_DBG("uri:[%s], oic_conn_type:[0x%x] To [%s:%d] ", uri, oic_conn_type, dev_addr.addr, dev_addr.port);
	ret = OCDoResource(NULL, method, uri, &dev_addr, (OCPayload *)payload, oic_conn_type,
			OC_HIGH_QOS, &cbdata, oic_options_ptr, options_size);
	free(uri);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResource(CRUD:%d) Fail(%d)", method, ret);
		OCRepPayloadDestroy(payload);
		_icl_ioty_free_response_container(cb_container);
		return ic_ioty_parse_oic_error(ret);
	}

	icl_get_timeout(&timeout);
	cb_container->timeout = g_timeout_add_seconds(timeout, _icl_ioty_response_timeout,
			cb_container);

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_get(iotcon_remote_resource_h resource,
		iotcon_query_h query, iotcon_remote_resource_response_cb cb, void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	ret = _icl_ioty_remote_resource_crud(IOTCON_REQUEST_GET, resource, NULL, query,
			cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_crud() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}


int icl_ioty_remote_resource_put(iotcon_remote_resource_h resource,
		iotcon_representation_h repr,
		iotcon_query_h query,
		iotcon_remote_resource_response_cb cb,
		void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	ret = _icl_ioty_remote_resource_crud(IOTCON_REQUEST_PUT, resource, repr, query,
			cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_crud() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_post(iotcon_remote_resource_h resource,
		iotcon_representation_h repr,
		iotcon_query_h query,
		iotcon_remote_resource_response_cb cb,
		void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	ret = _icl_ioty_remote_resource_crud(IOTCON_REQUEST_POST, resource, repr, query,
			cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_crud() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_delete(iotcon_remote_resource_h resource,
		iotcon_remote_resource_response_cb cb,
		void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	ret = _icl_ioty_remote_resource_crud(IOTCON_REQUEST_DELETE, resource, NULL, NULL,
			cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_crud() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}


static void _icl_ioty_encap_get_cb(iotcon_remote_resource_h resource,
		iotcon_error_e err,
		iotcon_request_type_e request_type,
		iotcon_response_h response,
		void *user_data)
{
	DBG(">>>>>>>> called... resource :[%p]", resource);

	int ret;
	iotcon_representation_h repr;
	iotcon_remote_resource_state_e state;

	RET_IF(NULL == resource);
	RET_IF(IOTCON_ERROR_NONE == err && NULL == response);

	/* MONITORING */
	if (resource->monitoring.presence) {
		switch (err) {
		case IOTCON_ERROR_NONE:
			state = (IOTCON_RESPONSE_OK == response->result) ?
				IOTCON_REMOTE_RESOURCE_ALIVE : IOTCON_REMOTE_RESOURCE_LOST_SIGNAL;
			break;
		default:
			state = IOTCON_REMOTE_RESOURCE_LOST_SIGNAL;
		}

		if (state != resource->monitoring.state) {
			DBG("call the monitoring callback [%p]", resource->monitoring.cb);
			resource->monitoring.cb(resource, state, resource->monitoring.user_data);
			resource->monitoring.state = state;
			DBG("call the monitoring callback done");
		}
	}

	/* CACHING */
	if (resource->caching.obs_handle) {
		switch (err) {
		case IOTCON_ERROR_NONE:
			repr = response->repr;
			break;
		default:
			WARN("_icl_ioty_encap_get_cb() Fail(%d)", err);
			repr = NULL;
		}

		ret = icl_representation_compare(resource->caching.repr, repr);
		if (IC_EQUAL != ret) { /* updated */
			if (resource->caching.repr)
				icl_representation_destroy(resource->caching.repr);
			resource->caching.repr = repr;
			if (response)
				response->repr = NULL;

			if (resource->caching.cb) {
				DBG("call the caching callback [%p]", resource->caching.cb);
				resource->caching.cb(resource, repr, resource->caching.user_data);
				DBG("call the caching callback done");
			}
		}
	}
}


static gboolean _icl_ioty_encap_timeout_cb(gpointer p)
{
	int ret;
	iotcon_remote_resource_h resource = p;

	RETV_IF(NULL == resource, G_SOURCE_REMOVE);

	/* get */
	ret = icl_ioty_remote_resource_get(resource, NULL, _icl_ioty_encap_get_cb, NULL);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_get() Fail(%d)", ret);
		return G_SOURCE_REMOVE;
	}

	/* get timeout */
	if (resource->timer_id)
		g_source_remove(resource->timer_id);

	resource->timer_id = g_timeout_add_seconds(resource->checking_interval,
			_icl_ioty_encap_timeout_cb, resource);

	return G_SOURCE_REMOVE;
}


int icl_ioty_remote_resource_set_checking_interval(iotcon_remote_resource_h resource,
		int time_interval)
{
	int ret;

	if (resource->timer_id)
		g_source_remove(resource->timer_id);

	ret = icl_ioty_remote_resource_get(resource, NULL, _icl_ioty_encap_get_cb, NULL);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_get() Fail(%d)", ret);
		return ret;
	}

	resource->timer_id = g_timeout_add_seconds(time_interval, _icl_ioty_encap_timeout_cb,
			resource);

	return IOTCON_ERROR_NONE;
}

/* LCOV_EXCL_START */
static void _icl_ioty_monitoring_presence_cb(iotcon_presence_h presence,
		iotcon_error_e err, iotcon_presence_response_h response, void *user_data)
{
	iotcon_remote_resource_h resource = user_data;

	RET_IF(NULL == resource);

	if (IOTCON_PRESENCE_OK == response->result
			&& IOTCON_PRESENCE_RESOURCE_DESTROYED != response->trigger)
		return;

	_icl_ioty_encap_timeout_cb(resource);
}
/* LCOV_EXCL_STOP */

int icl_ioty_remote_resource_start_monitoring(iotcon_remote_resource_h resource,
		iotcon_remote_resource_state_changed_cb cb, void *user_data)
{
	DBG(">>>>>>>> called... resource :[%p], cb :[%p]", resource, cb);
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	resource->monitoring.state = IOTCON_REMOTE_RESOURCE_ALIVE;
	resource->monitoring.cb = cb;
	resource->monitoring.user_data = user_data;

	/* presence */
	ret = icl_ioty_add_presence_cb(resource->host_address,
			resource->connectivity_type,
			NULL,
			_icl_ioty_monitoring_presence_cb,
			resource,
			&resource->monitoring.presence);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_add_presence_cb() Fail(%d)", ret);
		return ret;
	}

	/* get */
	if (resource->caching.obs_handle)
		return IOTCON_ERROR_NONE;

	ret = icl_ioty_remote_resource_set_checking_interval(resource,
			resource->checking_interval);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_set_checking_interval() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_stop_monitoring(iotcon_remote_resource_h resource)
{
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (NULL == resource->caching.obs_handle && resource->timer_id) {
		g_source_remove(resource->timer_id);
		resource->timer_id = 0;
	}

	ret = icl_ioty_remove_presence_cb(resource->monitoring.presence);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remove_presence_cb() Fail(%d)", ret);
		return ret;
	}
	resource->monitoring.presence = NULL;

	return IOTCON_ERROR_NONE;
}

static void _icl_ioty_caching_observe_cb(iotcon_remote_resource_h resource,
		iotcon_error_e err, int sequence_number, iotcon_response_h response, void *user_data)
{
	RET_IF(NULL == resource || NULL == response);


	if (resource->timer_id)
		g_source_remove(resource->timer_id);

	if (resource->caching.obs_handle) {
		int ret;
		iotcon_representation_h repr = response->repr;
		resource->timer_id = g_timeout_add_seconds(resource->checking_interval, _icl_ioty_encap_timeout_cb,
			resource);
		DBG("resource->timer_id :[%d]", resource->timer_id);

		ret = icl_representation_compare(resource->caching.repr, repr);
		if (IC_EQUAL != ret) { /* updated */
			if (resource->caching.repr)
				icl_representation_destroy(resource->caching.repr);
			resource->caching.repr = repr;
			if (response)
				response->repr = NULL;

			if (resource->caching.cb) {
				DBG("call the caching callback [%p]", resource->caching.cb);
				resource->caching.cb(resource, repr, resource->caching.user_data);
				DBG("call the caching callback done");
			}
		}
	}
}

int icl_ioty_remote_resource_start_caching(iotcon_remote_resource_h resource,
		iotcon_remote_resource_cached_representation_changed_cb cb, void *user_data)
{
	DBG(">>>>>>>> called... resource :[%p], cb :[%p]", resource, cb);
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	resource->caching.cb = cb;
	resource->caching.user_data = user_data;

	/* observe */
	ret = _icl_ioty_remote_resource_observe(resource, IOTCON_OBSERVE_IGNORE_OUT_OF_ORDER,
			NULL, _icl_ioty_caching_observe_cb, NULL, &(resource->caching.obs_handle));
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_remote_resource_observe() Fail(%d)", ret);
		return ret;
	}

	/* get */
	if (resource->monitoring.presence)
		return IOTCON_ERROR_NONE;

	DBG("checking_interval:[%d]", resource->checking_interval);

	ret = icl_ioty_remote_resource_set_checking_interval(resource,
			resource->checking_interval);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_set_checking_interval() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_remote_resource_stop_caching(iotcon_remote_resource_h resource)
{
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (NULL == resource->monitoring.presence && resource->timer_id) {
		g_source_remove(resource->timer_id);
		resource->timer_id = 0;
	}

	ret = icl_ioty_remote_resource_observe_cancel(resource,
			resource->caching.obs_handle);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_observe_cancel() Fail(%d)", ret);
		return IOTCON_ERROR_INVALID_PARAMETER;
	}
	resource->caching.obs_handle = NULL;

	return IOTCON_ERROR_NONE;
}


int icl_ioty_start_presence(unsigned int time_to_live)
{
	int ret;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}

	DBG("time_to_live:[%d]", time_to_live);
	ret = OCStartPresence(time_to_live);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCStartPresence() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_stop_presence()
{
	int ret;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCStopPresence();
	icl_ioty_mutex_unlock();
	if (OC_STACK_OK != ret) {
		ERR("OCStopPresence() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

static int _icl_ioty_resource_bind_type(OCResourceHandle handle,
		const char *resource_type)
{
	int ret;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCBindResourceTypeToResource(handle, resource_type);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCBindResourceTypeToResource() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

static int _icl_ioty_resource_bind_interface(OCResourceHandle handle,
		const char *iface)
{
	int ret;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCBindResourceInterfaceToResource(handle, iface);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCBindResourceInterfaceToResource() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_create(const char *uri_path,
		iotcon_resource_types_h res_types,
		iotcon_resource_interfaces_h ifaces,
		uint8_t policies,
		iotcon_request_handler_cb cb,
		void *user_data,
		iotcon_resource_h *resource_handle)
{
	GList *c;
	OCResourceHandle handle;
	int ret, i;
	iotcon_resource_h resource;

	RETV_IF(NULL == uri_path, IOTCON_ERROR_INVALID_PARAMETER);
	RETVM_IF(ICL_URI_PATH_LENGTH_MAX < strlen(uri_path),
			IOTCON_ERROR_INVALID_PARAMETER, "Invalid uri_path(%s)", uri_path);
	RETV_IF(NULL == res_types, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_handle, IOTCON_ERROR_INVALID_PARAMETER);

	resource = calloc(1, sizeof(struct icl_resource));
	if (NULL == resource) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	resource->uri_path = strdup(uri_path);
	resource->policies = policies;
	resource->types = icl_resource_types_ref(res_types);
	resource->ifaces = icl_resource_interfaces_ref(ifaces);

	resource->cb = cb;
	resource->user_data = user_data;

	policies = ic_ioty_convert_policies(policies);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		free(resource->uri_path);
		free(resource);
		return ret;
	}
	SECURE_DBG("res type [%s], res intf [%s], uri [%s], policy [0x%x]",
		(char *) res_types->type_list->data, (char *) ifaces->iface_list->data, uri_path, policies);
	ret = OCCreateResource(&handle, res_types->type_list->data, ifaces->iface_list->data,
			uri_path, icl_ioty_ocprocess_request_cb, resource, policies);
	icl_ioty_mutex_unlock();
	if (OC_STACK_OK != ret) {
		ERR("OCCreateResource() Fail(%d)", ret);
		free(resource->uri_path);
		free(resource);
		return IOTCON_ERROR_IOTIVITY;
	}
	resource->res_handle = handle;

	for (c = res_types->type_list, i = 0; c; c = c->next, i++) {
		if (0 == i) /* skip a first resource type */
			continue;
		_icl_ioty_resource_bind_type(handle, c->data);
	}

	for (c = ifaces->iface_list, i = 0; c; c = c->next, i++) {
		if (0 == i) /* skip a first resource iface */
			continue;
		_icl_ioty_resource_bind_interface(handle, c->data);
	}

	*resource_handle = resource;
	DBG("resource:[%p]", resource);

	if (NULL == resource_hash_table) {
		DBG("resource_hash_table will be created");
		resource_hash_table = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
	}
	g_hash_table_add(resource_hash_table, g_strdup_printf("%p", resource));

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_bind_type(iotcon_resource_h resource,
		const char *resource_type)
{
	FN_CALL;
	int ret;
	iotcon_resource_types_h resource_types;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_type, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_resource_types_clone(resource->types, &resource_types);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_resource_types_clone() Fail(%d)", ret);
		return ret;
	}

	ret = icl_resource_types_add(resource_types, resource_type);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_resource_types_add() Fail(%d)", ret);
		icl_resource_types_destroy(resource_types);
		return ret;
	}

	ret = _icl_ioty_resource_bind_type(resource->res_handle, resource_type);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_resource_bind_type Fail(%d)", ret);
		icl_resource_types_destroy(resource_types);
		return ret;
	}

	icl_resource_types_destroy(resource->types);
	resource->types = resource_types;

	return ret;
}

int icl_ioty_resource_bind_interface(iotcon_resource_h resource,
		const char *iface)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_resource_interfaces_add(resource->ifaces, iface);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_resource_interfaces_add() Fail(%d)", ret);
		return ret;
	}

	ret = _icl_ioty_resource_bind_interface(resource->res_handle, iface);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("_icl_ioty_resource_bind_interface() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_bind_child_resource(iotcon_resource_h parent,
		iotcon_resource_h child)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == parent, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == child, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(parent == child, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCBindResource(parent->res_handle, child->res_handle);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCBindResource() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_unbind_child_resource(iotcon_resource_h parent,
		iotcon_resource_h child)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == parent, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == child, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCUnBindResource(parent->res_handle, child->res_handle);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCUnBindResource() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_notify(iotcon_resource_h resource, iotcon_representation_h repr,
		iotcon_observers_h observers, iotcon_qos_e qos)
{
	int ret, obs_length = 0;
	GList *c;
	OCRepPayload *payload = NULL;
	iotcon_observers_h observers_ori;
	OCObservationId obs_ids[IC_OBSERVE_ID_MAX_LEN];
	OCQualityOfService oc_qos;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	observers_ori = (observers) ? observers : resource->observers;
	if (NULL == observers_ori) {
		DBG("observers_ori is NULL");
		return IOTCON_ERROR_NONE;
	}
	for (c = observers_ori->observers_list; c; c = c->next) {
		int observe_id = GPOINTER_TO_INT(c->data);
		obs_ids[obs_length++] = observe_id;
	}
	if (0 == obs_length) {
		DBG("No observers for lib");
		return IOTCON_ERROR_NONE;
	}

	ret = ic_ioty_convert_representation(repr, &payload);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("ic_ioty_convert_representation() Fail(%d)", ret);
		return ret;
	}

	oc_qos = ic_ioty_convert_qos(qos);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	if (payload) {
		ret = OCNotifyListOfObservers(resource->res_handle, obs_ids, obs_length, payload, oc_qos);
		OCRepPayloadDestroy(payload);
	} else {
		ret = OCNotifyAllObservers(resource->res_handle, oc_qos);
	}
	icl_ioty_mutex_unlock();

	if (OC_STACK_NO_OBSERVERS == ret) {
		WARN("No Observers. Stop Notifying");
		return IOTCON_ERROR_NONE;
	} else if (OC_STACK_OK != ret) {
		ERR("OCNotifyListOfObservers() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_resource_destroy(iotcon_resource_h resource)
{
	FN_CALL;
	int ret = IOTCON_ERROR_NONE;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (resource->res_handle) {
		ret = icl_ioty_mutex_lock();
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
			goto OUT;
		}
		ret = OCDeleteResource(resource->res_handle);
		icl_ioty_mutex_unlock();

		if (OC_STACK_OK != ret) {
			ERR("OCDeleteResource() Fail(%d)", ret);
			ret = ic_ioty_parse_oic_error(ret);
			goto OUT;
		}
	} else {
		ERR("resource->res_handle is NULL");
	}

OUT:
	DBG("resource:[%p]", resource);
	if (NULL != resource_hash_table) {
		gchar *key = NULL;
		key = g_strdup_printf("%p", resource);
		g_hash_table_remove(resource_hash_table, key);
		g_free(key);

		if (0 == g_hash_table_size(resource_hash_table)) {
			DBG("resource_hash_table will be destoryed");
			g_hash_table_destroy(resource_hash_table);
			resource_hash_table = NULL;
		}
	}

	free(resource->uri_path);
	free(resource);
	return ret;
}

int icl_ioty_lite_resource_create(const char *uri_path,
		iotcon_resource_types_h res_types,
		uint8_t policies,
		iotcon_attributes_h attributes,
		iotcon_lite_resource_post_request_cb cb,
		void *user_data,
		iotcon_lite_resource_h *resource_handle)
{

	FN_CALL;
	int ret, i;
	GList *c;
	iotcon_lite_resource_h resource;
	const char *res_iface = NULL;

	RETV_IF(NULL == uri_path, IOTCON_ERROR_INVALID_PARAMETER);
	RETVM_IF(ICL_URI_PATH_LENGTH_MAX < strlen(uri_path),
			IOTCON_ERROR_INVALID_PARAMETER, "Invalid uri_path(%s)", uri_path);
	RETV_IF(NULL == res_types, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_handle, IOTCON_ERROR_INVALID_PARAMETER);

	resource = calloc(1, sizeof(struct icl_lite_resource));
	if (NULL == resource) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}
	resource->uri_path = strdup(uri_path);
	resource->policies = policies;
	resource->policies = policies;
	resource->attributes = attributes;
	icl_attributes_ref(resource->attributes);
	resource->cb = cb;
	resource->cb_data = user_data;

	res_iface = IOTCON_INTERFACE_DEFAULT;

	policies = ic_ioty_convert_policies(policies);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		free(resource->uri_path);
		free(resource);
		return ret;
	}

	SECURE_DBG("res type [%s], uri [%s], policy [0x%x]",
		(char *) res_types->type_list->data, uri_path, policies);
	ret = OCCreateResource(&(resource->res_handle), res_types->type_list->data, res_iface, uri_path,
			icl_ioty_ocprocess_lite_request_cb, resource, policies);
	icl_ioty_mutex_unlock();
	if (OC_STACK_OK != ret) {
		ERR("OCCreateResource() Fail(%d)", ret);
		free(resource->uri_path);
		free(resource);
		return IOTCON_ERROR_IOTIVITY;
	}

	for (c = res_types->type_list, i = 0; c; c = c->next, i++) {
		if (0 == i) /* skip a first resource type */
			continue;
		_icl_ioty_resource_bind_type(resource->res_handle, c->data);
	}

	*resource_handle = resource;

	return IOTCON_ERROR_NONE;
}

int icl_ioty_lite_resource_destroy(iotcon_lite_resource_h resource)
{
	FN_CALL;
	int ret = IOTCON_ERROR_NONE;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (resource->res_handle) {
		ret = icl_ioty_mutex_lock();
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
			goto OUT;
		}
		ret = OCDeleteResource(resource->res_handle);
		icl_ioty_mutex_unlock();

		if (OC_STACK_OK != ret) {
			ERR("OCDeleteResource() Fail(%d)", ret);
			ret = ic_ioty_parse_oic_error(ret);
			goto OUT;
		}
	} else {
		ERR("resource->res_handle is NULL");
	}

OUT:
	free(resource->uri_path);
	free(resource);
	return ret;
}

int icl_ioty_lite_resource_notify(iotcon_lite_resource_h resource)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCNotifyAllObservers(resource->res_handle, IOTCON_QOS_HIGH);
	icl_ioty_mutex_unlock();

	if (OC_STACK_NO_OBSERVERS == ret) {
		WARN("No Observers. Stop Notifying");
		return IOTCON_ERROR_NONE;
	} else if (OC_STACK_OK != ret) {
		ERR("OCNotifyListOfObservers() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	return IOTCON_ERROR_NONE;
}

int icl_ioty_lite_resource_update_attributes(iotcon_lite_resource_h resource,
		iotcon_attributes_h attributes)
{
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (attributes)
		attributes = icl_attributes_ref(attributes);

	if (resource->attributes)
		icl_attributes_destroy(resource->attributes);

	resource->attributes = attributes;

	icl_ioty_lite_resource_notify(resource);

	return IOTCON_ERROR_NONE;

}

int icl_ioty_response_send(iotcon_response_h response_handle)
{
	FN_CALL;
	int ret;
	OCEntityHandlerResponse response = {0};
	OCRepPayload *payload = NULL;

	RETV_IF(NULL == response_handle, IOTCON_ERROR_INVALID_PARAMETER);

	response.requestHandle = response_handle->oic_request_h;
	response.resourceHandle = response_handle->oic_resource_h;
	response.ehResult = ic_ioty_convert_response_result(response_handle->result);

	if (response_handle->header_options && response_handle->header_options->hash) {
		ic_ioty_convert_header_options(response_handle->header_options,
				response.sendVendorSpecificHeaderOptions,
				sizeof(response.sendVendorSpecificHeaderOptions)
				/ sizeof(response.sendVendorSpecificHeaderOptions[0]));

		response.numSendVendorSpecificHeaderOptions =
			g_hash_table_size(response_handle->header_options->hash);
	}


	ic_ioty_convert_representation(response_handle->repr, &payload);
	response.payload = (OCPayload*)payload;

	/* related to block transfer */
	response.persistentBufferFlag = 0;

	ret = icl_ioty_mutex_lock();
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_mutex_lock() Fail(%d)", ret);
		return ret;
	}
	ret = OCDoResponse(&response);
	OCRepPayloadDestroy(payload);
	icl_ioty_mutex_unlock();

	if (OC_STACK_OK != ret) {
		ERR("OCDoResponse() Fail(%d)", ret);
		return ic_ioty_parse_oic_error(ret);
	}

	FN_END;
	return IOTCON_ERROR_NONE;
}

