/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "iotcon-types.h"
#include "ic.h"
#include "ic-utils.h"
#include "ic-options.h"
#include "ic-representation.h"
#include "ic-remote-resource.h"
#include "ic-resource.h"
#include "ic-resource-types.h"
#include "ic-resource-interfaces.h"
#include "ic-ioty.h"

#define ICL_REMOTE_RESOURCE_DEFAULT_CHECKING_INTERVAL 10 /* 10 sec */

/* The length of resource_type should be less than or equal to 61.
 * If resource_type is NULL, then All resources in host are discovered. */
int icl_find_resource(const char *host_address,
		int connectivity_type,
		iotcon_query_h query,
		iotcon_found_resource_cb cb,
		void *user_data)
{
	int ret;

	RETV_IF(false == ic_utils_check_connectivity_type(connectivity_type),
			IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	ret = icl_ioty_find_resource(host_address, connectivity_type, query, cb,
			user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_find_resource() Fail(%d)", ret);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}

/* If you know the information of resource, then you can make a proxy of the resource. */
int icl_remote_resource_create(const char *host_address,
		iotcon_connectivity_type_e connectivity_type,
		const char *uri_path,
		uint8_t policies,
		iotcon_resource_types_h resource_types,
		iotcon_resource_interfaces_h resource_ifaces,
		iotcon_remote_resource_h *resource_handle)
{
	FN_CALL;
	char temp[PATH_MAX] = {0};
	iotcon_remote_resource_h resource = NULL;

	RETV_IF(NULL == host_address, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == uri_path, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(false == icl_resource_check_uri_path(uri_path),
			IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_types, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_ifaces, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == resource_handle, IOTCON_ERROR_INVALID_PARAMETER);
	RETVM_IF(IOTCON_CONNECTIVITY_ALL == connectivity_type, IOTCON_ERROR_INVALID_PARAMETER,
		"Should use specific connectivity type of the remote resource");

	if (IC_EQUAL == strncmp(IC_COAP_PREFIX, host_address, strlen(IC_COAP_PREFIX)))
		snprintf(temp, sizeof(temp), "%s", host_address);
	else if (policies & IOTCON_RESOURCE_SECURE)
		snprintf(temp, sizeof(temp), "%s%s", IC_COAPS, host_address);
	else
		snprintf(temp, sizeof(temp), "%s%s", IC_COAP, host_address);

	resource = calloc(1, sizeof(struct icl_remote_resource));
	if (NULL == resource) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	resource->connectivity_options = ic_utils_host_address_get_connectivity(temp,
			connectivity_type);

	resource->connectivity_type = connectivity_type;
	resource->host_address = ic_utils_strdup(temp);
	resource->uri_path = ic_utils_strdup(uri_path);
	resource->policies = policies;
	resource->types = icl_resource_types_ref(resource_types);
	resource->ifaces = icl_resource_interfaces_ref(resource_ifaces);
	resource->checking_interval = ICL_REMOTE_RESOURCE_DEFAULT_CHECKING_INTERVAL;
	resource->ref_count = 1;

	*resource_handle = resource;

	DBG("remote resource : [%p]", resource);

	return IOTCON_ERROR_NONE;
}

void icl_remote_resource_ref(iotcon_remote_resource_h resource)
{
	RET_IF(NULL == resource);

	resource->ref_count++;
}

void icl_remote_resource_unref(iotcon_remote_resource_h resource)
{
	RET_IF(NULL == resource);

	resource->ref_count--;
	if (0 != resource->ref_count)
		return;

	/* ref_count is zero, so destroy remote resource */
	DBG("remote resource will be freed.[%p]", resource);
	if (true == resource->is_found) {
		ERR("It can't be destroyed by user.");
		return;
	}

	free(resource->uri_path);
	free(resource->host_address);
	free(resource->device_id);
	free(resource->device_name);
	icl_resource_interfaces_destroy(resource->ifaces);
	icl_resource_types_destroy(resource->types);

	/* null COULD be allowed */
	if (resource->header_options)
		icl_options_destroy(resource->header_options);

	if (resource->monitoring.presence)
		icl_remote_resource_stop_monitoring(resource);

	if (resource->caching.obs_handle)
		icl_remote_resource_stop_caching(resource);

	free(resource);
}


int icl_remote_resource_destroy(iotcon_remote_resource_h resource)
{
	FN_CALL;
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if ((NULL != resource->obs_handle) && (false == resource->is_found))
		icl_ioty_remote_resource_observe_deregister(resource);

	icl_remote_resource_unref(resource);

	return IOTCON_ERROR_NONE;
}

static bool _icl_remote_resource_header_foreach_cb(unsigned short id,
		const char *data, void *user_data)
{
	int ret;
	iotcon_remote_resource_h resource = user_data;

	RETV_IF(NULL == resource, IOTCON_FUNC_STOP);

	if (NULL == resource->header_options) {
		ret = icl_options_create(&resource->header_options);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("resource->header_options() Fail(%d)", ret);
			return IOTCON_FUNC_STOP;
		}
	}

	ret = icl_options_add(resource->header_options, id, data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_options_add() Fail(%d)", ret);
		return IOTCON_FUNC_STOP;
	}

	return IOTCON_FUNC_CONTINUE;
}

int icl_remote_resource_clone(iotcon_remote_resource_h src,
		iotcon_remote_resource_h *dest)
{
	FN_CALL;
	int ret;
	iotcon_remote_resource_h resource = NULL;

	RETV_IF(NULL == src, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == dest, IOTCON_ERROR_INVALID_PARAMETER);

	resource = calloc(1, sizeof(struct icl_remote_resource));
	if (NULL == resource) {
		ERR("calloc() Fail(%d)", errno);
		return IOTCON_ERROR_OUT_OF_MEMORY;
	}

	resource->uri_path = ic_utils_strdup(src->uri_path);
	resource->host_address = ic_utils_strdup(src->host_address);
	resource->connectivity_type = src->connectivity_type;
	resource->connectivity_options = src->connectivity_options;
	resource->device_id = ic_utils_strdup(src->device_id);
	resource->device_name = ic_utils_strdup(src->device_name);
	resource->policies = src->policies;
	resource->checking_interval = src->checking_interval;
	resource->ref_count = 1;

	if (src->header_options) {
		ret = icl_options_foreach(src->header_options,
				_icl_remote_resource_header_foreach_cb, resource);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("iotcon_options_foreach() Fail(%d)", ret);
			icl_remote_resource_destroy(resource);
			return ret;
		}
	}

	ret = icl_resource_types_clone(src->types, &resource->types);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_resource_types_clone() Fail(%d)", ret);
		icl_remote_resource_destroy(resource);
		return ret;
	}

	ret = icl_resource_interfaces_clone(src->ifaces, &resource->ifaces);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_resource_interfaces_clone() Fail(%d)", ret);
		icl_remote_resource_destroy(resource);
		return ret;
	}

	resource->connectivity_type = src->connectivity_type;

	*dest = resource;
	DBG("remote resource : [%p]", resource);

	return IOTCON_ERROR_NONE;
}

int icl_remote_resource_start_caching(iotcon_remote_resource_h resource,
		iotcon_remote_resource_cached_representation_changed_cb cb, void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (true == resource->is_found) {
		ERR("The resource should be cloned.");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}
	if (resource->caching.obs_handle) {
		ERR("Already Start Caching");
		return IOTCON_ERROR_ALREADY;
	}

	icl_remote_resource_ref(resource);
	ret = icl_ioty_remote_resource_start_caching(resource, cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_start_caching() Fail(%d)", ret);
		icl_remote_resource_unref(resource);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}


int icl_remote_resource_stop_caching(iotcon_remote_resource_h resource)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (0 == resource->caching.obs_handle) {
		ERR("Not Cached");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	ret = icl_ioty_remote_resource_stop_caching(resource);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_stop_caching() Fail(%d)", ret);
		return ret;
	}
	icl_remote_resource_unref(resource);

	return IOTCON_ERROR_NONE;
}

int icl_remote_resource_start_monitoring(
		iotcon_remote_resource_h resource,
		iotcon_remote_resource_state_changed_cb cb,
		void *user_data)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	if (true == resource->is_found) {
		ERR("The resource should be cloned.");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}
	if (resource->monitoring.presence) {
		ERR("Already Start Monitoring");
		return IOTCON_ERROR_ALREADY;
	}

	resource->monitoring.state = IOTCON_REMOTE_RESOURCE_ALIVE;

	icl_remote_resource_ref(resource);
	ret = icl_ioty_remote_resource_start_monitoring(resource, cb, user_data);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_start_monitoring() Fail(%d)", ret);
		icl_remote_resource_unref(resource);
		return ret;
	}

	return IOTCON_ERROR_NONE;
}


int icl_remote_resource_stop_monitoring(iotcon_remote_resource_h resource)
{
	FN_CALL;
	int ret;

	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (NULL == resource->monitoring.presence) {
		ERR("Not Monitoring");
		return IOTCON_ERROR_INVALID_PARAMETER;
	}

	ret = icl_ioty_remote_resource_stop_monitoring(resource);
	if (IOTCON_ERROR_NONE != ret) {
		ERR("icl_ioty_remote_resource_stop_monitoring() Fail(%d)", ret);
		return ret;
	}
	icl_remote_resource_unref(resource);

	return IOTCON_ERROR_NONE;
}
