/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <glib.h>
#include <glib-object.h>

#include "ic.h"
#include "ic-utils.h"
#include "ic-ioty.h"

#define ICL_TIMEOUT_MAX 60*60 /* 60 min */
#define ICL_TIMEOUT_DEFAULT 30 /* 30 sec */
#define ICL_DEFAULT_DEVICE_NAME "UNKNOWN"

#define ICL_POLLING_INTERVAL_MAX 999 /* 999 ms */

#define ICL_IOTY_TIME_INTERVAL_DEFAULT 100 /* 100 ms */

static pthread_t icl_thread;
static int icl_timeout_seconds = ICL_TIMEOUT_DEFAULT;
static int icl_init_count;

static int icl_ioty_polling_interval = ICL_IOTY_TIME_INTERVAL_DEFAULT;

int icl_initialize(const char *file_path, bool is_pt)
{
	int ret;

	RETV_IF(NULL == file_path, IOTCON_ERROR_INVALID_PARAMETER);

#if !GLIB_CHECK_VERSION(2, 35, 0)
	g_type_init();
#endif

	ic_utils_mutex_lock(IC_UTILS_MUTEX_INIT);
	if (0 == icl_init_count) {
		ret = icl_ioty_set_persistent_storage(file_path, is_pt);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_set_persistent_storage() Fail(%d)", ret);
			return ret;
		}

		ret = icl_ioty_init(&icl_thread);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_init() Fail(%d)", ret);
			ic_utils_mutex_unlock(IC_UTILS_MUTEX_INIT);
			return ret;
		}

		ret = icl_ioty_set_platform_info();
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_set_platform_info() Fail(%d)", ret);
			icl_ioty_deinit(icl_thread);
			ic_utils_mutex_unlock(IC_UTILS_MUTEX_INIT);
			return ret;
		}

		ret = icl_ioty_set_device_info(ICL_DEFAULT_DEVICE_NAME);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_set_device_info() Fail(%d)", ret);
			icl_ioty_deinit(icl_thread);
			ic_utils_mutex_unlock(IC_UTILS_MUTEX_INIT);
			return ret;
		}
	}
	icl_init_count++;
	ic_utils_mutex_unlock(IC_UTILS_MUTEX_INIT);

	return IOTCON_ERROR_NONE;
}

int icl_deinitialize(void)
{
	ic_utils_mutex_lock(IC_UTILS_MUTEX_INIT);

	icl_init_count--;
	if (0 == icl_init_count) {
		icl_ioty_deinit(icl_thread);
		icl_thread = 0;
	}

	ic_utils_mutex_unlock(IC_UTILS_MUTEX_INIT);

	return IOTCON_ERROR_NONE;
}

int icl_get_timeout(int *timeout_seconds)
{
	RETV_IF(NULL == timeout_seconds, IOTCON_ERROR_INVALID_PARAMETER);

	*timeout_seconds = icl_timeout_seconds;

	return IOTCON_ERROR_NONE;
}

int icl_set_timeout(int timeout_seconds)
{
	if (ICL_TIMEOUT_MAX < timeout_seconds || timeout_seconds <= 0) {
		ERR("timeout_seconds(%d) must be in range from 1 to 3600", timeout_seconds);
		return IOTCON_ERROR_INVALID_PARAMETER;
	}
	INFO("timeout_seconds [%d]", timeout_seconds);

	icl_timeout_seconds = timeout_seconds;

	return IOTCON_ERROR_NONE;
}

int icl_polling_get_interval(int *interval)
{
	RETV_IF(NULL == interval, IOTCON_ERROR_INVALID_PARAMETER);

	*interval = icl_ioty_polling_interval;

	return IOTCON_ERROR_NONE;
}

int icl_polling_set_interval(int interval)
{
	RETV_IF(interval <= 0, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(interval > ICL_POLLING_INTERVAL_MAX, IOTCON_ERROR_INVALID_PARAMETER);

	INFO("interval [%d]", interval);

	icl_ioty_polling_interval = interval;

	ic_utils_cond_signal(IC_UTILS_COND_POLLING);

	return IOTCON_ERROR_NONE;
}

int icl_polling_invoke(void)
{
	ic_utils_cond_signal(IC_UTILS_COND_POLLING);

	return IOTCON_ERROR_NONE;
}
