/* Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdlib.h>
#include <errno.h>
#include <glib.h>

#include "iotcon-types.h"
#include "ic.h"
#include "ic-list.h"
#include "ic-value.h"
#include "ic-utils.h"
#include "ic-representation.h"
#include "ic-attributes.h"

API int iotcon_attributes_create(iotcon_attributes_h *ret_attributes)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_create(ret_attributes);
}


API int iotcon_attributes_destroy(iotcon_attributes_h attributes)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_destroy(attributes);
}

API int iotcon_attributes_remove(iotcon_attributes_h attributes, const char *key)
{
	API_CALL;
	gboolean ret = FALSE;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup(%s) Fail", key);
		return IOTCON_ERROR_NO_DATA;
	}

	ret = g_hash_table_remove(attributes->hash_table, key);
	if (FALSE == ret) {
		ERR("g_hash_table_remove(%s) Fail", key);
		return IOTCON_ERROR_NO_DATA;
	}

	return IOTCON_ERROR_NONE;
}


API int iotcon_attributes_get_int(iotcon_attributes_h attributes, const char *key,
		int *val)
{
	API_CALL;
	iotcon_value_h value;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	icl_basic_s *real = (icl_basic_s*)value;
	if (IOTCON_TYPE_INT != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*val = real->val.i;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_int(iotcon_attributes_h attributes, const char *key,
		int val)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_int(attributes, key, val);
}

API int iotcon_attributes_get_bool(iotcon_attributes_h attributes, const char *key,
		bool *val)
{
	API_CALL;
	icl_basic_s *real = NULL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_basic_s*)value;
	if (IOTCON_TYPE_BOOL != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*val = real->val.b;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_bool(iotcon_attributes_h attributes, const char *key,
		bool val)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_bool(attributes, key, val);
}

API int iotcon_attributes_get_double(iotcon_attributes_h attributes,
		const char *key, double *val)
{
	API_CALL;
	icl_basic_s *real = NULL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_basic_s*)value;
	if (IOTCON_TYPE_DOUBLE != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*val = real->val.d;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_double(iotcon_attributes_h attributes,
		const char *key, double val)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_double(attributes, key, val);
}

API int iotcon_attributes_get_str(iotcon_attributes_h attributes, const char *key,
		char **val)
{
	API_CALL;
	icl_basic_s *real = NULL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_basic_s*)value;
	if (IOTCON_TYPE_STR != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*val = real->val.s;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_str(iotcon_attributes_h attributes, const char *key,
		char *val)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_str(attributes, key, val);
}

API int iotcon_attributes_get_byte_str(iotcon_attributes_h attributes, const char *key,
		unsigned char **val, int *len)
{
	API_CALL;
	iotcon_value_h value = NULL;
	icl_val_byte_str_s *real = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == val, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == len, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_val_byte_str_s*)value;
	if (IOTCON_TYPE_BYTE_STR != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*val = real->s;
	*len = real->len;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_byte_str(iotcon_attributes_h attributes,
		const char *key, unsigned char *val, int len)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_byte_str(attributes, key, val, len);
}

API int iotcon_attributes_is_null(iotcon_attributes_h attributes, const char *key,
		bool *is_null)
{
	API_CALL;
	icl_basic_s *real = NULL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == is_null, IOTCON_ERROR_INVALID_PARAMETER);

	value = (iotcon_value_h) g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_basic_s*)value;
	*is_null = (IOTCON_TYPE_NULL == real->type) ? true : false;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_null(iotcon_attributes_h attributes, const char *key)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_null(attributes, key);
}

API int iotcon_attributes_get_list(iotcon_attributes_h attributes, const char *key,
		iotcon_list_h *list)
{
	API_CALL;
	iotcon_value_h value = NULL;
	icl_val_list_s *real = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == list, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_val_list_s*)value;
	if (IOTCON_TYPE_LIST != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*list = real->list;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_list(iotcon_attributes_h attributes, const char *key,
		iotcon_list_h list)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_list(attributes, key, list);
}

API int iotcon_attributes_get_attributes(iotcon_attributes_h src, const char *key,
		iotcon_attributes_h *dest)
{
	API_CALL;
	icl_val_attributes_s *real = NULL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == src, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == dest, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(src->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}

	real = (icl_val_attributes_s*)value;
	if (IOTCON_TYPE_ATTRIBUTES != real->type) {
		ERR("Invalid Type(%d)", real->type);
		return IOTCON_ERROR_INVALID_TYPE;
	}

	*dest = real->attributes;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_add_attributes(iotcon_attributes_h attributes,
		const char *key, iotcon_attributes_h val)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_add_attributes(attributes, key, val);
}

API int iotcon_attributes_get_type(iotcon_attributes_h attributes, const char *key,
		iotcon_type_e *type)
{
	API_CALL;
	iotcon_value_h value = NULL;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == key, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == type, IOTCON_ERROR_INVALID_PARAMETER);

	value = g_hash_table_lookup(attributes->hash_table, key);
	if (NULL == value) {
		ERR("g_hash_table_lookup() Fail");
		return IOTCON_ERROR_NO_DATA;
	}
	*type = value->type;

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_get_keys_count(iotcon_attributes_h attributes,
		unsigned int *count)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == count, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == attributes->hash_table, IOTCON_ERROR_INVALID_PARAMETER);

	*count = g_hash_table_size(attributes->hash_table);

	return IOTCON_ERROR_NONE;
}

API int iotcon_attributes_clone(iotcon_attributes_h attributes,
		iotcon_attributes_h *attributes_clone)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);

	return icl_attributes_clone(attributes, attributes_clone);
}

API int iotcon_attributes_foreach(iotcon_attributes_h attributes,
		iotcon_attributes_cb cb, void *user_data)
{
	API_CALL;
	GHashTableIter iter;
	gpointer key;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == attributes, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == cb, IOTCON_ERROR_INVALID_PARAMETER);

	g_hash_table_iter_init(&iter, attributes->hash_table);
	while (g_hash_table_iter_next(&iter, &key, NULL)) {
		if (IOTCON_FUNC_STOP == cb(attributes, key, user_data))
			break;
	}

	return IOTCON_ERROR_NONE;
}

