/*
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "iotcon-types.h"
#include "ic.h"
#include "ic-utils.h"
#include "ic-options.h"
#include "ic-representation.h"
#include "ic-remote-resource.h"
#include "ic-resource.h"
#include "ic-resource-types.h"
#include "ic-resource-interfaces.h"
#include "ic-ioty.h"

#define ICL_REMOTE_RESOURCE_MAX_CHECKING_INTERVAL 3600 /* 60 min */

/* The length of resource_type should be less than or equal to 61.
 * If resource_type is NULL, then All resources in host are discovered. */
API int iotcon_find_resource(const char *host_address,
		int connectivity_type,
		iotcon_query_h query,
		iotcon_found_resource_cb cb,
		void *user_data)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(false == ic_utils_check_permission(IC_PERMISSION_INTERNET),
			IOTCON_ERROR_PERMISSION_DENIED);
	return icl_find_resource(host_address, connectivity_type, query, cb, user_data);
}

/* If you know the information of resource, then you can make a proxy of the resource. */
API int iotcon_remote_resource_create(const char *host_address,
		iotcon_connectivity_type_e connectivity_type,
		const char *uri_path,
		uint8_t policies,
		iotcon_resource_types_h resource_types,
		iotcon_resource_interfaces_h resource_ifaces,
		iotcon_remote_resource_h *resource_handle)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	return icl_remote_resource_create(host_address, connectivity_type, uri_path,
						policies, resource_types, resource_ifaces, resource_handle);

}

API int iotcon_remote_resource_destroy(iotcon_remote_resource_h resource)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	return icl_remote_resource_destroy(resource);
}

API int iotcon_remote_resource_clone(iotcon_remote_resource_h src,
		iotcon_remote_resource_h *dest)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	return icl_remote_resource_clone(src, dest);
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_uri_path(iotcon_remote_resource_h resource,
		char **uri_path)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == uri_path, IOTCON_ERROR_INVALID_PARAMETER);

	*uri_path = resource->uri_path;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_host_address(
		iotcon_remote_resource_h resource, char **host_address)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == host_address, IOTCON_ERROR_INVALID_PARAMETER);

	*host_address = resource->host_address;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_connectivity_type(
		iotcon_remote_resource_h resource, iotcon_connectivity_type_e *connectivity_type)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == connectivity_type, IOTCON_ERROR_INVALID_PARAMETER);

	*connectivity_type = resource->connectivity_type;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_device_id(iotcon_remote_resource_h resource,
		char **device_id)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == device_id, IOTCON_ERROR_INVALID_PARAMETER);
	RETVM_IF(NULL == resource->device_id, IOTCON_ERROR_NO_DATA,
			"If you want to get device ID, you should call iotcon_find_resource().");

	*device_id = resource->device_id;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_device_name(iotcon_remote_resource_h resource,
		char **device_name)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == device_name, IOTCON_ERROR_INVALID_PARAMETER);
	RETVM_IF(NULL == resource->device_name, IOTCON_ERROR_NO_DATA,
			"If you want to get device name, you should call iotcon_find_resource().");

	*device_name = resource->device_name;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_types(iotcon_remote_resource_h resource,
		iotcon_resource_types_h *types)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == types, IOTCON_ERROR_INVALID_PARAMETER);

	*types = resource->types;

	return IOTCON_ERROR_NONE;
}


/* The content of the resource should not be freed by user. */
API int iotcon_remote_resource_get_interfaces(iotcon_remote_resource_h resource,
		iotcon_resource_interfaces_h *ifaces)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == ifaces, IOTCON_ERROR_INVALID_PARAMETER);

	*ifaces = resource->ifaces;

	return IOTCON_ERROR_NONE;
}


API int iotcon_remote_resource_get_policies(iotcon_remote_resource_h resource,
		uint8_t *policies)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == policies, IOTCON_ERROR_INVALID_PARAMETER);

	*policies = resource->policies;

	return IOTCON_ERROR_NONE;
}

API int iotcon_remote_resource_get_options(iotcon_remote_resource_h resource,
		iotcon_options_h *options)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == options, IOTCON_ERROR_INVALID_PARAMETER);
	WARN_IF(NULL == resource->header_options, "Not Set header options");

	*options = resource->header_options;

	return IOTCON_ERROR_NONE;
}

/* if header_options is NULL, then client's header_options is unset */
API int iotcon_remote_resource_set_options(iotcon_remote_resource_h resource,
		iotcon_options_h options)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);

	if (options)
		options = icl_options_ref(options);

	if (resource->header_options)
		icl_options_destroy(resource->header_options);

	resource->header_options = options;

	return IOTCON_ERROR_NONE;
}

API int iotcon_remote_resource_get_checking_interval(iotcon_remote_resource_h resource,
		int *interval)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == interval, IOTCON_ERROR_INVALID_PARAMETER);

	*interval = resource->checking_interval;

	return IOTCON_ERROR_NONE;
}


API int iotcon_remote_resource_set_checking_interval(iotcon_remote_resource_h resource,
		int interval)
{
	API_CALL;
	int ret;

	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(ICL_REMOTE_RESOURCE_MAX_CHECKING_INTERVAL < interval || interval <= 0,
			IOTCON_ERROR_INVALID_PARAMETER);

	if (resource->monitoring.presence || resource->caching.obs_handle) {
		ret = icl_ioty_remote_resource_set_checking_interval(resource, interval);
		if (IOTCON_ERROR_NONE != ret) {
			ERR("icl_ioty_remote_resource_set_checking_interval() Fail(%d)", ret);
			return ret;
		}
	}

	resource->checking_interval = interval;

	return IOTCON_ERROR_NONE;
}


/*
 *******************
 * start / stop caching
 *******************
 */
API int iotcon_remote_resource_start_caching(iotcon_remote_resource_h resource,
		iotcon_remote_resource_cached_representation_changed_cb cb, void *user_data)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(false == ic_utils_check_permission(IC_PERMISSION_INTERNET),
			IOTCON_ERROR_PERMISSION_DENIED);
	return icl_remote_resource_start_caching(resource, cb, user_data);
}


API int iotcon_remote_resource_stop_caching(iotcon_remote_resource_h resource)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(false == ic_utils_check_permission(IC_PERMISSION_INTERNET),
			IOTCON_ERROR_PERMISSION_DENIED);
	return icl_remote_resource_stop_caching(resource);
}


API int iotcon_remote_resource_get_cached_representation(
		iotcon_remote_resource_h resource,
		iotcon_representation_h *representation)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(NULL == resource, IOTCON_ERROR_INVALID_PARAMETER);
	RETV_IF(NULL == representation, IOTCON_ERROR_INVALID_PARAMETER);
	WARN_IF(NULL == resource->caching.repr, "No Cached Representation");

	*representation = resource->caching.repr;

	return IOTCON_ERROR_NONE;
}


/*
 *******************
 * start / stop monitoring
 *******************
 */
API int iotcon_remote_resource_start_monitoring(
		iotcon_remote_resource_h resource,
		iotcon_remote_resource_state_changed_cb cb,
		void *user_data)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(false == ic_utils_check_permission(IC_PERMISSION_INTERNET),
			IOTCON_ERROR_PERMISSION_DENIED);
	return icl_remote_resource_start_monitoring(resource, cb, user_data);
}


API int iotcon_remote_resource_stop_monitoring(iotcon_remote_resource_h resource)
{
	API_CALL;
	RETV_IF(false == ic_utils_check_ocf_feature(), IOTCON_ERROR_NOT_SUPPORTED);
	RETV_IF(false == ic_utils_check_permission(IC_PERMISSION_INTERNET),
			IOTCON_ERROR_PERMISSION_DENIED);
	return icl_remote_resource_stop_monitoring(resource);
}

